<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE rss PUBLIC "-//Netscape Communications//DTD RSS 0.91//EN" "http://www.rssboard.org/rss-0.91.dtd">
<rss version="0.91">
  <channel>
    <title>{{ channel_name }} updates</title>
    <link>{{ channel_url }}</link>
    <description>Updates in the last two weeks</description>
    <language>en</language>
    <copyright>Copyright {{ current_time | strftime("%Y") }}, Anaconda, Inc.</copyright>
    <pubDate>{{ current_time | strftime("%a, %d %b %Y %H:%M:%S %z") }}</pubDate>
{% for name in package_order %}
    <item>
      <title>{{ name }} {{ channeldata_packages[name].version }}</title>
      <pubDate>{{ package_mtimes[name] | strftime("%a, %d %b %Y %H:%M:%S %z") }}</pubDate>
    </item>
{%- endfor %}
  </channel>
</rss>